/*
 * Kommunikation per Message Queue.
 */

#include "fpconfig.h"

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <fpmsgqueue.h>
#include "fpmode.h"

char * fpmsgqueue_id = "$Id: fpmsgqueue.C,v 1.1.1.1 1997/10/28 10:49:43 frank Exp $";

/*
 * Leerer Konstruktor
 */

fpMsgQueue::fpMsgQueue (long id)
{
  msgId = -1;
  myId  =  id;
}

/*
 * Copy Constructor
 */

fpMsgQueue::fpMsgQueue (const fpMsgQueue &other)
{
  msgId = -1;
  myId  = -1;

  if (other.msgId != -1) {
    attach (other.msgKey);
  }
}

/*
 * Full Constructor
 */

fpMsgQueue::fpMsgQueue (key_t key, int mode, int excl)
{
  msgId = -1;
  myId  = -1;

  create (key, mode, excl);
}

/*
 * Message Queue anzapfen oder selber erzeugen
 */

int
fpMsgQueue::create (key_t key, int mode, int excl)
{
  msgId  = -1;
  msgKey = key;

  /*
   * wenn mode != 0 versuche, die Message Queue explizit anzulegen
   */

  if (mode != 0) {
    msgId = msgget (key, IPC_CREAT | IPC_EXCL | mode);

    if (msgId != -1) {
      arefirst = 1;
      seterror ();
      return 0;
    }

    /*
     * Gut, unser Versuch, die Message Queue exklusiv einzurichten, ist
     * fehlgeschlagen. Wahrscheinlich existiert die Resource bereits.
     */

    if (errno != EEXIST) {
      /*
       * Nein, das ist nicht das Problem.
       */
      seterror (errno);
      return -1;
    }

    if (excl) {
      /* heul, wir wollten aber selber initialisieren! */
      seterror (EEXIST);
      return -1;
    }
  }

  /*
   * Okay, haengen wir uns an die existierende Message Queue mit dran.
   */

  if ((msgId = msgget (key, 0)) == -1) {
    seterror (errno);
    return -1;
  }

  arefirst = 0;
  seterror ();
  return 0;
}

/*
 * Destruktor.
 */

fpMsgQueue::~fpMsgQueue ()
{
}

/*
 * Weitere Verwaltung
 */

fpMsgQueue &
fpMsgQueue::operator= (const fpMsgQueue &other)
{
  arefirst =  0;
  msgId    = -1;
  myId     = -1;
  seterror ();

  if (other.msgId != -1) {
    attach (other.msgKey);
  }

  return *this;
}

int
fpMsgQueue::destroy (void)
{
  if (msgId == -1) {
    seterror ();
    return 0;
  }

  if (msgctl (msgId, IPC_RMID, NULL) == -1) {
    seterror (errno);
    return -1;
  }

  msgId   = -1;
  seterror ();
  return 0;
}

/*
 * Operationen
 */

int
fpMsgQueue::size (void)
{
  struct msqid_ds buf;

  if (msgctl (msgId, IPC_STAT, &buf) == -1) {
    seterror (errno);
    return 0;
  }

  return (int) buf.msg_qbytes;
}

int
fpMsgQueue::send (void *data, int length, long dest, int block)
{
  char *packet;
  int result;

  if ((data == NULL && length > 0) || length < 0) {
    seterror (EINVAL);
    return -1;
  }

  if ((packet = (char *) malloc (length + 2*sizeof(long))) == NULL) {
    seterror (ENOMEM);
    return -1;
  }

  ((long *) packet)[0] = dest;
  ((long *) packet)[1] = getId ();

  if (length) {
    memcpy (packet + 2*sizeof(long), data, length);
  }

  result = msgsnd (packet, length+sizeof(long), (block)?0:IPC_NOWAIT);

  free (packet);
  return result;
}

int
fpMsgQueue::recv (void *data, int length, long *src, int block)
{
  char *packet;
  int result;

  if ((data == NULL && length > 0) || length < 0) {
    seterror (EINVAL);
    return -1;
  }

  if ((packet = (char *) malloc (length + 2*sizeof(long))) == NULL) {
    seterror (ENOMEM);
    return -1;
  }

  result = msgrcv (packet, length+sizeof(long), getId(),
		   (block) ? 0 : IPC_NOWAIT);

  if (result >= 0) {
    result -= sizeof(long);
    if (result > 0) {
      memcpy (data, packet+2*sizeof(long), result);
    }

    if (src != NULL) {
      *src = ((long *) packet)[1];
    }
  }

  free (packet);
  return result;
}

int
fpMsgQueue::send (fpMsgPacket *packet, long dest, int block)
{
  int result;

  if (packet == NULL) {
    seterror (EINVAL);
    return -1;
  }

  packet->setSource (getId());
  packet->setDest   (dest);

  result = msgsnd (packet->getPkgData(),
		   packet->getPkgSize(),
		   (block) ? 0 : IPC_NOWAIT);

  return result;
}

int
fpMsgQueue::recv (fpMsgPacket *packet, int block)
{
  int result;

  if (packet == NULL) {
    seterror (EINVAL);
    return -1;
  }

  result = msgrcv (packet->getPkgData(),
		   packet->getPkgSize(),
		   getId (),
		   (block) ? 0 : IPC_NOWAIT);

  return result;
}
