/*
 * Implementation einer etwas komfortableren Message-Queue, April 96 __FP__
 *
 * $Id: fpmsgqueue.h,v 1.2 1997/12/09 12:30:24 frank Exp $
 */

#ifndef __FPMSGQUEUE_H__
#define __FPMSGQUEUE_H__

#include "fpconfig.h"

#include <sys/types.h>
#include <sys/ipc.h>
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#include <sys/msg.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <errno.h>
#include "ipcbase.h"
#include "fpmode.h"

/*
 * ----------------------------------------------------------------------
 * Hilfsklasse zum Verschicken beliebiger Datenpakete
 * ----------------------------------------------------------------------
 */

struct fpMsgPacket {
  virtual void   setSource  (long)       = 0;
  virtual void   setDest    (long)       = 0;
  virtual int    getPkgSize (void) const = 0;
  virtual void * getPkgData (void) const = 0;
};

template <class T>
struct fpPkgStruct {
  long dest;
  long src;
  T    data;
};

template <class T>
class fpPacket : public fpMsgPacket {
protected:
  void   setSource  (long);
  void   setDest    (long);
  int    getPkgSize (void) const;
  void * getPkgData (void) const;

public:
  T *    getData    (void) const;
  void   setData    (const T *);
  long   source     (void) const;
  operator T*();
  T & operator= (const T &);

private:
  fpPkgStruct<T> data;
};

/*
 * ----------------------------------------------------------------------
 * Message Queue Klasse
 * ----------------------------------------------------------------------
 */

class fpMsgQueue : virtual public cIPC {
public:
  fpMsgQueue (key_t, int = 0, int = 0);
  fpMsgQueue (const fpMsgQueue &);
  fpMsgQueue (long = -1);

  /*
   * Verwaltung
   */

  int attach  (const fpMsgQueue *);
  int attach  (key_t);
  int create  (key_t, int = S_IRWXU, int = 0);
  int detach  (void);
  int destroy (void);

  fpMsgQueue & operator= (const fpMsgQueue &);

  /*
   * Eindeutige ID
   */

  void setId (long = -1);
  long getId (void) const;

  /*
   * Senden und Empfangen, System-Stil
   */

  int msgsnd (void *, int, int);
  int msgrcv (void *, int, long, int);

  /*
   * Senden und Empfangen von Rohdaten
   */

  int send (void *, int, long,   int=1);
  int recv (void *, int, long *, int=1);

  /*
   * Senden und Empfangen fertig geschnuerter Pakete
   */

  int send  (fpMsgPacket *, long, int=1);
  int recv  (fpMsgPacket *, int=1);

  /*
   * Operationen
   */

  int first  (void) const;
  int size   (void);

  virtual ~fpMsgQueue ();

protected:
  key_t  getkey (void) const;

private:
  int   arefirst;
  key_t msgKey;
  int   msgId;
  long  myId;
};

inline int
fpMsgQueue::attach (key_t key)
{
  return create (key, 0);
}

inline int
fpMsgQueue::attach (const fpMsgQueue * other)
{
  return attach (other->msgKey);
}

inline int
fpMsgQueue::detach (void)
{
  msgId   = -1;
  seterror ();
  return 0;
}

inline int
fpMsgQueue::first (void) const
{
  return arefirst;
}

inline int
fpMsgQueue::msgsnd (void *data, int msgsz, int msgflg)
{
  int res = ::msgsnd (msgId, (struct msgbuf *) data, msgsz, msgflg);
  seterror (errno);
  return res;
}

inline int
fpMsgQueue::msgrcv (void *data, int msgsz, long msgtyp, int msgflg)
{
  int res = ::msgrcv (msgId, (struct msgbuf *) data, msgsz, msgtyp, msgflg);
  seterror (errno);
  return res;
}

inline void
fpMsgQueue::setId (long id)
{
  myId = id;
}

inline long
fpMsgQueue::getId (void) const
{
  if (myId == -1) {
    return (long) getpid ();
  }
  return myId;
}

inline key_t
fpMsgQueue::getkey (void) const
{
  if (msgId == -1) return 0;
  return msgKey;
}

/*
 * ----------------------------------------------------------------------
 * Implementation der Template-Klasse
 * ----------------------------------------------------------------------
 */

template <class T>
inline void
fpPacket<T>::setSource (long src)
{
  data.src = src;
}

template <class T>
inline void
fpPacket<T>::setDest (long dest)
{
  data.dest = dest;
}

/*
 * Paketgroesse, wie wir sie mit msgsnd verwenden wollen
 */

template <class T>
inline int
fpPacket<T>::getPkgSize (void) const
{
  return sizeof (fpPkgStruct<T>) - sizeof (long);
}

template <class T>
void *
fpPacket<T>::getPkgData (void) const
{
  return (void *) &data;
}

template <class T>
T *
fpPacket<T>::getData (void) const
{
  return & data.data;
}

template <class T>
void
fpPacket<T>::setData (const T *other)
{
  data.data = *other;
}

template <class T>
long
fpPacket<T>::source (void) const
{
  return data.src;
}

template <class T>
fpPacket<T>::operator T*()
{
  return & data.data;
}

template <class T>
T &
fpPacket<T>::operator= (const T &other)
{
  data.data = other;
  return data.data;
}

#endif
