/*
 * Praktische Semaphoren-Klasse
 */

#include "fpconfig.h"

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <stdlib.h>
#include <errno.h>
#include <memory.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <fpsema.h>
#include "ipcbase.h"
#include "fpmode.h"

#ifdef NEED_UNION_SEMUN
union semun {
  int val;
  struct semid_ds *buf;
  u_short *array;
};
#endif

char * fpsema_id = "$Id: fpsema.C,v 1.1.1.1 1997/10/28 10:49:44 frank Exp $";

/*
 * Leerer Konstruktor
 */

fpSema::fpSema (void)
{
  semId = -1;
}

/*
 * Copy Constructor
 */

fpSema::fpSema (const fpSema &other)
  : semId (-1)
{
  if (other.semId != -1) {
    attach (other.semKey);
  }
}

/*
 * Kompletter Konstruktor
 */

fpSema::fpSema (key_t key, int nsems, int mode, int excl,
		const u_short *vals)
  : semId (-1)
{
  create (key, nsems, mode, excl, vals);
}

/*
 * Leerer Destruktor, kann von anderen Klassen ueberladen werden.
 */

fpSema::~fpSema ()
{
  /* this place intentionally left blank */
}

/*
 * Erzeuge Segment oder haenge dich an existierendes dran
 */

int
fpSema::create (key_t key, int nsems, int mode, int excl,
		const u_short *vals)
{
  struct semid_ds seminfo;
  struct sembuf semops;
  union semun semunion;
  u_short *array;
  int count;

  if (semId != -1) {
    seterror (EEXIST);
    return -1;
  }

  semKey = key;
  seterror ();

  /*
   * Wenn nsems>0, versuche, die Semaphore explizit anzulegen
   */

  if (nsems > 0 && mode != 0) {
    semId = semget (key, nsems, IPC_CREAT | IPC_EXCL | mode);

    if (semId != -1) {
      /*
       * Okay, wir sind die ersten
       */

      arefirst = 1;

      /*
       * Initialisiere die Semaphoren. Zum Schluss fuehren wir eine
       * beliebige semop durch, um semid_ds.sem_otime zu setzen, was
       * anderen Prozessen die gelungene Initialisierung anzeigt.
       */

      if ((array = new u_short[nsems]) == NULL) {
	/* oops */
	semctl (semId, 0, IPC_RMID, semunion);
	seterror (ENOMEM);
	semId = -1;
	return -1;
      }

      if (vals != NULL) {
	for (count=0; count<nsems; count++)
	  array[count] = vals[count];
	array[0] += 1; /* das regeln wir gleich per semop */
      }
      else {
	for (count=1; count<nsems; count++)
	  array[count] = 0;
	array[0] = 2;
      }

      semunion.array = array;
      if (semctl (semId, 0, SETALL, semunion) == -1) {
	semctl (semId, 0, IPC_RMID, semunion);
	seterror (errno);
	delete [] array;
	return -1;
      }
      delete [] array;

      /*
       * jetzt fuehren wir eine semop durch (eben haben wir array[0]
       * um 1 zu hoch initialisiert). Das signalisert durch Setzen von
       * semid_ds.sem_otime anderen Prozessen, dass die Initialisierung
       * abgeschlossen ist.
       * Die Operation kann prinzipiell nicht schiefgehen.
       */

      semops.sem_num =  0;
      semops.sem_op  = -1;
      semops.sem_flg = IPC_NOWAIT;

      if (::semop (semId, &semops, 1) == -1) {
	semctl (semId, 0, IPC_RMID, semunion);
	seterror (errno);
	semId = -1;
	return -1;
      }

      /*
       * Fertig!
       */

      seterror ();
      return 0;
    }

    /*
     * Gut, unser Versuch, die Semaphore exklusiv einzurichten, ist
     * fehlgeschlagen. Wahrscheinlich existiert die Resource bereits.
     */

    if (errno != EEXIST) {
      /*
       * Nein, das ist nicht das Problem.
       */
      seterror (errno);
      return -1;
    }

    if (excl) {
      /* heul, wir wollten aber selber initialisieren! */
      seterror (EEXIST);
      return -1;
    }
  }

  arefirst = 0;

  /*
   * Okay, haengen wir uns an die existierende Semaphoren mit dran.
   * Beachte, dass der Aufruf fehlschlaegt, wenn das existierende Set
   * weniger als nsems Semaphoren hat.
   */

  if ((semId = semget (key, nsems, 0)) == -1) {
    seterror (errno);
    return -1;
  }

  /*
   * Gut. Teste, ob die Initialisierung bereits abgeschlossen ist.
   */

  semunion.buf = &seminfo;

  if (semctl (semId, 0, IPC_STAT, semunion) == -1) {
    seterror (errno);
    semId = -1;
    return -1;
  }

  if (seminfo.sem_otime == 0) {
    /*
     * Nein, die Initialisierung der Semaphoren ist noch nicht abgeschlossen.
     * Wir wollen nicht warten, deshalb melden wir einfach einen EEXIST.
     */
    seterror (EEXIST);
    semId = -1;
    return -1;
  }

  /*
   * Sehr gut.
   */

  seterror (0);
  return 0;
}

/*
 * Operations
 */

int
fpSema::destroy (void)
{
  union semun dummy;

  if (semId == -1) {
    seterror ();
    return 0;
  }

  if (semctl (semId, 0, IPC_RMID, dummy) == -1) {
    if (errno == EIDRM) {
      semId = -1;
    }
    seterror (errno);
    return -1;
  }

  seterror ();
  semId = -1;
  return 0;
}

int
fpSema::semop (struct sembuf *sops, int nops)
{
  if (nops < 0) {
    seterror (EINVAL);
    return -1;
  }
  if (::semop (semId, sops, nops) == -1) {
    seterror (errno);
    return -1;
  }
  seterror ();
  return 0;
}

int
fpSema::size (void)
{
  struct semid_ds seminfo;
  union semun semunion;

  semunion.buf = &seminfo;

  if (semctl (semId, 0, IPC_STAT, semunion) == -1) {
    seterror (errno);
    return -1;
  }

  return (int) seminfo.sem_nsems;
}

fpSema &
fpSema::operator= (const fpSema &other)
{
  arefirst =  0;
  semId    = -1;
  seterror ();

  if (other.semId != -1) {
    attach (other.semKey);
  }

  return *this;
}

/*
 * ----------------------------------------------------------------------
 * Implementation der Operationen via fpSemaHelper
 * ----------------------------------------------------------------------
 */

int
fpSemaHelper::operator= (int val)
{
  union semun dummy;

  if (id == -1) {
    master->errorset (EINVAL);
    return -1;
  }

  dummy.val = val;

  if (semctl (id, sem, SETVAL, dummy) == -1) {
    master->errorset (errno);
    return -1;
  }

  master->errorset (0);
  return 0;
}

fpSemaHelper::operator int() const
{
  union semun dummy;
  int res;

  if (id == -1) {
    master->errorset (EINVAL);
    return -1;
  }

  if ((res = semctl (id, sem, GETVAL, dummy)) == -1) {
    master->errorset (errno);
    return -1;
  }

  master->errorset (0);
  return res;
}

int
fpSemaHelper::doop (int op, int block)
{
  struct sembuf sops;

  sops.sem_num = sem;
  sops.sem_op  = op;
  sops.sem_flg = (block) ? 0 : IPC_NOWAIT;

  if (::semop (id, &sops, 1) == -1) {
    master->errorset (errno);
    return -1;
  }
  master->errorset (0);
  return 0;
}
