/*
 * Implementation einer etwas komfortableren Semaphoren-Klasse
 *
 * $Id: fpsema.h,v 1.3 1997/10/28 11:50:25 frank Exp $
 */

#ifndef __FPSEMA_H__
#define __FPSEMA_H__

#include "fpconfig.h"

#include <sys/types.h>
#include <sys/ipc.h>
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#include <stdlib.h>

#include "ipcbase.h"
#include "fpmode.h"

/*
 * ----------------------------------------------------------------------
 * Hilfsklasse um "Arithmetik" mit Konstrukten der Form fpSema[i]
 * zu ermoeglichen
 * ----------------------------------------------------------------------
 */

class fpSema;

class fpSemaHelper {
public:
  fpSemaHelper & set (int, int, class fpSema *);

  int operator=  (int);
  int operator++ (void);	/* increment sema, shouldn't fail */
  int operator++ (int);		/* (postfix ++)                   */
  int operator-- (void);	/* decrement sema, might block    */
  int operator-- (int);		/* (postfix --)                   */
  int operator+= (int);		/* add to sema, shouldn't fail    */
  int operator-= (int);		/* sub from sema, might block     */
  int incr       (int=1);	/* ++ with or without blocking    */
  int decr       (int=1);	/* -- with or without blocking    */
  int zero       (int=1);	/* blocks until sema becomes 0    */
  int add        (int, int=1);	/* += with or without blocking    */
  int sub        (int, int=1);	/* -= with or without blocking    */
  operator int   () const;	/* get semaphore value            */

protected:
  int doop       (int, int);	/* perform operation on sema      */
  
private:
  int id, sem;
  fpSema *master;
};

/*
 * ----------------------------------------------------------------------
 * Semaphoren-Klasse
 * ----------------------------------------------------------------------
 */

class fpSema : virtual public cIPC {
public:
  fpSema (key_t, int=0, int=0, int=0, const u_short * = NULL);
  fpSema (const fpSema &);
  fpSema (void);

  /*
   * Verwaltung
   */

  int attach  (const fpSema *);
  int attach  (key_t, int = 0);
  int create  (key_t, int, int = S_IRWXU, int = 0, const u_short * = NULL);
  int detach  (void);
  int destroy (void);

  fpSema & operator= (const fpSema &);

  /*
   * Operationen
   */

  int semop  (struct sembuf *, int);
  int first  (void) const;	/* were we the first?       */
  int size   (void);            /* number of sems in set    */

  /*
   * Helfer
   */

  void errorset (int = 0);
  fpSemaHelper &operator[] (int);

  virtual ~fpSema ();

protected:
  key_t  getkey (void) const;

private:
  key_t  semKey;
  int    semId;
  int    arefirst;
  fpSemaHelper helper;
};

/*
 * ----------------------------------------------------------------------
 * Inline-Funktionen der Hilfsklasse
 * ----------------------------------------------------------------------
 */

inline fpSemaHelper &
fpSemaHelper::set (int theid, int thesem, class fpSema *themaster)
{
  id     = theid;
  sem    = thesem;
  master = themaster;

  return *this;
}

/*
 * map all operations to doop
 */

inline int fpSemaHelper::operator++ (int n) { (void) n; return doop ( 1, 1); }
inline int fpSemaHelper::operator++ (void)  { return doop ( 1, 1); }
inline int fpSemaHelper::operator-- (int n) { (void) n; return doop (-1, 1); }
inline int fpSemaHelper::operator-- (void)  { return doop (-1, 1); }
inline int fpSemaHelper::operator+= (int x) { return doop ( x, 1); }
inline int fpSemaHelper::operator-= (int x) { return doop (-x, 1); }
inline int fpSemaHelper::incr       (int b) { return doop ( 1, b); }
inline int fpSemaHelper::decr       (int b) { return doop (-1, b); }
inline int fpSemaHelper::zero       (int b) { return doop ( 0, b); }

/*
 * add and sub must be guarded against the special meaning
 * of 0 in a semaphore operation
 */

inline int 
fpSemaHelper::add (int x, int b)
{
  if (x==0) return 0;
  return doop ( x, b);
}
inline int
fpSemaHelper::sub (int x, int b)
{ 
  if (x==0) return 0;
  return doop (-x, b);
}

/*
 * ----------------------------------------------------------------------
 * Inline-Funktionen fpSema
 * ----------------------------------------------------------------------
 */

inline void
fpSema::errorset (int err)
{
  cIPC::seterror (err);
}

inline int
fpSema::first (void) const
{
  return arefirst;
}

inline fpSemaHelper &
fpSema::operator[] (int i)
{
  return helper.set (semId, i, this);
}

inline int
fpSema::attach (key_t key, int nsems)
{
  return create (key, nsems, 0, 0, NULL);
}

inline int
fpSema::attach (const fpSema * other)
{
  return attach (other->semKey);
}

inline int
fpSema::detach (void)
{
  seterror ();
  semId = -1;
  return 0;
}

inline key_t
fpSema::getkey (void) const
{
  if (semId == -1) return 0;
  return semKey;
}

#endif
